==> pyodps-0.3.4.tar.gz: pyodps-0.3.4/odps/df/backends/optimize.py <==
#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import itertools

from .core import Backend
from ..expr.expressions import *
from ..expr.groupby import GroupByCollectionExpr, SequenceGroupBy
from ..expr.reduction import SequenceReduction
from ..expr.collections import DistinctCollectionExpr
from ..expr.utils import get_attrs


class Optimizer(Backend):
    def __init__(self, expr, memo=None):
        self._expr = expr
        self._memo = dict() if memo is None else memo

    def optimize(self):
        for node in self._expr.traverse(parent_cache=self._memo, top_down=True,
                                        unique=True):
            try:
                node.accept(self)
            except NotImplementedError:
                continue

        return self._memo.get(id(self._expr)) or self._expr

    def _get_parent(self, expr):
        return self._memo.get(id(expr))

    def _sub(self, expr, to_sub):
        parents = self._get_parent(expr)
        if parents is None:
            self._expr = to_sub
        else:
            [p.substitute(expr, to_sub, parent_cache=self._memo)
             for p in set(parents)]

    def visit_filter_collection(self, expr):
        if isinstance(expr.input, GroupByCollectionExpr):
            predicate = self._broadcast_field(expr.predicate, expr.input)

            having = expr.input.having
            if having is not None:
                predicates = having & predicate
            else:
                predicates = predicate

            attrs = get_attrs(expr.input)
            attrs_values = dict((attr, getattr(expr.input, attr)) for attr in attrs)
            groupby_collection = type(expr.input)(**attrs_values)
            groupby_collection._having = predicates

            self._sub(expr, groupby_collection)

        raise NotImplementedError

    def visit_project_collection(self, expr):
        # Summary does not attend here
        self._visit_need_compact_collection(expr)

    def visit_groupby(self, expr):
        self._visit_need_compact_collection(expr)

    def visit_distinct(self, expr):
        self._visit_need_compact_collection(expr)

    def _visit_need_compact_collection(self, expr):
        compacted = self._compact(expr)
        if compacted is None:
            return

        self._sub(expr, compacted)

    def _add_parent_memo(self, expr, parent):
        if id(expr) not in self._memo:
            self._memo[id(expr)] = set([parent, ])
        else:
            self._memo[id(expr)].add(parent)

    def _compact(self, expr):
        to_compact = [expr, ]

        for node in expr.traverse(top_down=True, unique=True):
            if node is expr:
                continue
            if not isinstance(node, CollectionExpr):
                continue

            if isinstance(node, ProjectCollectionExpr) and \
                    not node.optimize_banned:
                valid = True
                for it in itertools.chain(*(node.all_path(to_compact[-1]))):
                    if isinstance(it, SequenceReduction):
                        valid = False
                        break
                if not valid:
                    break
                to_compact.append(node)
            else:
                break

        if len(to_compact) <= 1:
            return

        changed = False
        for field in self._get_fields(expr):
            self._add_parent_memo(field, expr)

            if not isinstance(field, SequenceExpr):
                continue
            broadcast_field = self._broadcast_field(field, to_compact[-1])
            if broadcast_field is not None:
                changed = True
                expr.substitute(field, broadcast_field, parent_cache=self._memo)

        if changed:
            expr.substitute(expr.input, to_compact[-1].input,
                            parent_cache=self._memo)
            return expr

    def _broadcast_field(self, expr, collection):
        changed = False
        for path in expr.all_path(collection, strict=True):
            col = None
            for i, node in enumerate(path):
                if isinstance(node, Column):
                    col = node
                    if i > 0:
                        self._add_parent_memo(col, path[i - 1])
                    continue
                elif isinstance(node, SequenceExpr):
                    continue
                elif isinstance(node, Scalar):
                    break

                # find a collection
                if col is not None:
                    field = self._get_field(node, col.source_name or col.name)

                    parents = set(self._memo[id(col)])
                    for parent in parents:
                        parent.substitute(col, field, parent_cache=self._memo)
                        self._add_parent_memo(field, parent)

                    col = field
                else:
                    if i > 0:
                        parent = path[i - 1]
                        parent.substitute(node, collection.input)
                        self._add_parent_memo(collection.input, parent)

                changed = True

        if changed:
            return expr

    def _get_fields(self, collection):
        if isinstance(collection, (ProjectCollectionExpr, Summary)):
            return collection.fields
        elif isinstance(collection, DistinctCollectionExpr):
            return collection.unique_fields
        elif isinstance(collection, GroupByCollectionExpr):
            return collection.fields

    def _get_field(self, collection, name):
        idx = collection.schema._name_indexes[name.lower()]
        return self._get_fields(collection)[idx]

==> pyodps-0.4.1.tar.gz: pyodps-0.4.1/odps/df/backends/optimize.py <==
#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import itertools

from .core import Backend
from ..expr.expressions import *
from ..expr.groupby import GroupByCollectionExpr, GroupbyAppliedCollectionExpr
from ..expr.reduction import SequenceReduction
from ..expr.collections import DistinctCollectionExpr, RowAppliedCollectionExpr
from ..expr.utils import get_attrs
from ...models import Schema


class Optimizer(Backend):
    def __init__(self, expr, traversed=None, use_cache=None):
        self._expr = expr
        self._traversed = traversed or set()
        self._use_cache = use_cache if use_cache is not None \
            else options.df.use_cache

    def optimize(self):
        for node in self._expr.traverse(top_down=True, unique=True,
                                        traversed=self._traversed):
            expr = self._handle_cache(node, root=self._expr)
            if node is self._expr and expr is not None:
                break
            if expr is not None:
                continue
            try:
                node.accept(self)
            except NotImplementedError:
                continue

        return self._expr

    def _handle_cache(self, expr, root=None):
        if not self._use_cache:
            return

        if expr is root and expr._cache_data is not None:
            sub = None
            if isinstance(expr, CollectionExpr):
                sub = CollectionExpr(_source_data=expr._cache_data,
                                      _schema=expr._schema)
            elif isinstance(expr, SequenceExpr):
                if expr is not root:
                    return
                sub = CollectionExpr(_source_data=expr._cache_data,
                                     _schema=Schema.from_lists([expr.name], [expr.dtype]))
                sub = sub[expr.name]
            elif isinstance(expr, Scalar):
                sub = Scalar(_value=expr._cache_data, _value_type=expr.dtype)

            if sub:
                if expr is root:
                    parents = None
                else:
                    parents = expr.parents
                self._sub(expr, sub, parents)
                return sub

    def _sub(self, expr, to_sub, parents):
        if not parents and expr is self._expr:
            self._expr = to_sub
        else:
            [p.substitute(expr, to_sub) for p in parents]

    def visit_filter_collection(self, expr):
        if isinstance(expr.input, GroupByCollectionExpr):
            predicate = self._broadcast_field(expr.predicate, expr.input)

            having = expr.input.having
            if having is not None:
                predicates = having & predicate
            else:
                predicates = predicate

            parents = expr.parents

            attrs = get_attrs(expr.input)
            attrs_values = dict((attr, getattr(expr.input, attr)) for attr in attrs)
            groupby_collection = type(expr.input)(**attrs_values)
            groupby_collection._having = predicates

            self._sub(expr, groupby_collection, parents)

        raise NotImplementedError

    def visit_project_collection(self, expr):
        # Summary does not attend here
        if not options.df.optimize:
            return
        self._visit_need_compact_collection(expr)

    def visit_groupby(self, expr):
        if not options.df.optimize:
            return
        self._visit_need_compact_collection(expr)

    def visit_distinct(self, expr):
        if not options.df.optimize:
            return
        self._visit_need_compact_collection(expr)

    def visit_apply_collection(self, expr):
        if not options.df.optimize:
            return
        self._visit_need_compact_collection(expr)

    def _visit_need_compact_collection(self, expr):
        parents = expr.parents

        compacted = self._compact(expr)
        if compacted is None:
            return

        self._sub(expr, compacted, parents)

    def _compact(self, expr):
        to_compact = [expr, ]

        for node in expr.traverse(top_down=True, unique=True):
            if node is expr:
                continue
            if not isinstance(node, CollectionExpr):
                continue

            if isinstance(node, ProjectCollectionExpr) and \
                    not node.optimize_banned:
                valid = True
                for it in itertools.chain(*(node.all_path(to_compact[-1]))):
                    if isinstance(it, SequenceReduction):
                        valid = False
                        break
                if not valid:
                    break
                to_compact.append(node)
            else:
                break

        if len(to_compact) <= 1:
            return

        changed = False
        for field in self._get_fields(expr):
            if not isinstance(field, SequenceExpr):
                continue
            broadcast_field = self._broadcast_field(field, to_compact[-1])
            if broadcast_field is not None:
                changed = True
                expr.substitute(field, broadcast_field)

        if changed:
            expr.substitute(expr.input, to_compact[-1].input)
            return expr

    def _broadcast_field(self, expr, collection):
        changed = False
        for path in expr.all_path(collection, strict=True):
            col = None
            for i, node in enumerate(path):
                if isinstance(node, Column):
                    col = node
                    continue
                elif isinstance(node, SequenceExpr):
                    continue
                elif isinstance(node, Scalar):
                    break

                # find a collection
                if col is not None:
                    parents = col.parents
                    field = self._get_field(node, col.source_name or col.name)
                    if col.is_renamed():
                        field = field.rename(col.name)
                    self._sub(col, field, parents)

                    col = next(n for n in field.traverse(top_down=True, unique=True)
                               if isinstance(n, Column))
                else:
                    if i > 0:
                        parent = path[i - 1]
                        parent.substitute(node, collection.input)

                changed = True

        if changed:
            return expr

    def _get_fields(self, collection):
        if isinstance(collection, (ProjectCollectionExpr, Summary)):
            return collection.fields
        elif isinstance(collection, DistinctCollectionExpr):
            return collection.unique_fields
        elif isinstance(collection, GroupByCollectionExpr):
            return collection.fields
        elif isinstance(collection, (RowAppliedCollectionExpr, GroupbyAppliedCollectionExpr)):
            return collection.fields

    def _get_field(self, collection, name):
        idx = collection.schema._name_indexes[name.lower()]
        return self._get_fields(collection)[idx]
